#include <iostream>
#include <assert.h>
#include <cilk/cilk.h>
#include <cilk/cilk_api.h>
#include <time.h>
using namespace std;

long print_time(std::ostream &out, struct timespec start,
                       struct timespec end){
    time_t s = end.tv_sec - start.tv_sec;
    long ns = end.tv_nsec - start.tv_nsec;
    if (ns < 0) {
		ns += 1000000000;
		s -= 1;
    }
    return (s * 1000 + ((ns + 500000) / 1000000));
}


/* mm_loop_serial is the naive iterative method */
void mm_loop_serial1(int* C, int* A, int* B, int n)
{
/* DO NOT MODIFY THIS CODE. THIS IS USED TO VERIFY THAT YOUR
* CODE IS PRODUCING THE RIGHT ANSWER. */
int i,j,k;


for (i=0; i<n; ++i){
  for (j=0; j<n; ++j){
    for (k=0; k<n; ++k){
      // printf("%d  ",i*n+j);
      C[i*n+j] = C[i*n+j] + (A[i*n+k] * B[k*n+j]);    //C[i*n+j] = A[n*i+j] + A[n*i+j]  ;
     }
   }
}


}
/* mm_loop_parallel is the parallel implementation of mm_loop_serial */
void mm_loop_serial2(int* C, int k0, int k1, int* A, int i0, int i1, int* B, int j0, int j1,  int n)
{
/* MODIFY THIS CODE TO MAKE IT PARALLEL */ 
	for (int i=i0; i<i1; i++){
		for (int j=j0; j<j1; j++)
			for (int k=k0; k<k1; k++)
				C[i*n+j] += A[i*n+k] * B[k*n+j];
}
 
}
/* random_matrix creates a random n-by-n matrix */
void random_matrix(int *A, int n)
{
	for(int i=0;i<n*n;i++) 
	{
  		A[i] = rand()%n;

	}
}
/* Print an n-by-n matrix */
void print_matrix(int *a, int n)
{
	for(int i=0;i<n;i++) 
	{
		for(int j=0; j<n; j++) 
		{
			printf("%d ", a[n*i+j]);
			if (j == n-1) printf("\n");
		}
	}

	printf("\n");
}

void serial_dandc(int i0, int i1, int j0, int j1, int k0, int k1, int* A, int lda, int* B, int ldb, int* C, int ldc, int X)

{
    int di = i1 - i0;
    int dj = j1 - j0;
    int dk = k1 - k0;
    if (di >= dj && di >= dk && di >= X) {
        int mi = i0 + di / 2;
        serial_dandc(i0, mi, j0, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
        serial_dandc(mi, i1, j0, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
  } else if (dj >= dk && dj >= X) {

        int mj = j0 + dj / 2;
        serial_dandc(i0, i1, j0, mj, k0, k1, A, lda, B, ldb, C, ldc,X);
        serial_dandc(i0, i1, mj, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
    } else if (dk >= X) {    
   
        int mk = k0 + dk / 2;
        serial_dandc(i0, i1, j0, j1, k0, mk, A, lda, B, ldb, C, ldc,X);
        serial_dandc(i0, i1, j0, j1, mk, k1, A, lda, B, ldb, C, ldc,X);
      

    } else {
        mm_loop_serial2(C, k0, k1,  A, i0, i1, B, j0, j1, lda)  ;
       /* for (int i = i0; i < i1; ++i)
            for (int j = j0; j < j1; ++j)
                for (int k = k0; k < k1; ++k)
                    C[i * ldc + j] += A[i * lda + k] * B[k * ldb + j];*/
    }
}

void parallel_dandc(int i0, int i1, int j0, int j1, int k0, int k1, int* A, int lda, int* B, int ldb, int* C, int ldc, int X)

{
    int di = i1 - i0;
    int dj = j1 - j0;
    int dk = k1 - k0;
    if (di >= dj && di >= dk && di >= X) {
        int mi = i0 + di / 2;
    cilk_spawn parallel_dandc(i0, mi, j0, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
        parallel_dandc(mi, i1, j0, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
    cilk_sync;
  } else if (dj >= dk && dj >= X) {

        int mj = j0 + dj / 2;
     cilk_spawn   parallel_dandc(i0, i1, j0, mj, k0, k1, A, lda, B, ldb, C, ldc,X);
        parallel_dandc(i0, i1, mj, j1, k0, k1, A, lda, B, ldb, C, ldc,X);
     cilk_sync;
    } else if (dk >= X) {    
   
        int mk = k0 + dk / 2;
        parallel_dandc(i0, i1, j0, j1, k0, mk, A, lda, B, ldb, C, ldc,X);
        parallel_dandc(i0, i1, j0, j1, mk, k1, A, lda, B, ldb, C, ldc,X);

    } else {
        mm_loop_serial2(C, k0, k1,  A, i0, i1, B, j0, j1, lda)  ;
       /* for (int i = i0; i < i1; ++i)
            for (int j = j0; j < j1; ++j)
                for (int k = k0; k < k1; ++k)
                    C[i * ldc + j] += A[i * lda + k] * B[k * ldb + j];*/
    }
}

int main(int argc , char * argv[]) 
{
	int n, s,X;
	int* a;
	int* b;
	int* c;
	int* d;
	if (argc == 1)
        {
                cout<<"Usage: ./DnC_MM order basecase"<<endl;
                exit(0);
        }
	else if ( argc > 2 )
        {
               // __cilkrts_set_param("nworkers", argv[1]);
		n = atoi(argv[1]);
		X = atoi(argv[2]);
        }
	s = n * n;
	if (s < 1000000000) 
	{
		a = (int *) malloc(s * sizeof(int));
		random_matrix(a,n);
		b = (int *) malloc(s * sizeof(int));
		random_matrix(b,n);
		if (n < 10) 
			print_matrix(a,n);
		if (n < 10) 
			print_matrix(b,n);
		c = (int *) malloc(s * sizeof(int));
		d = (int *) malloc(s * sizeof(int));
		for(int i=0;i<n;i++) 
		{
   			for(int j=0;j<n;j++) 
			{
    				c[i*n + j] = 0;
    				d[i*n +j] = 0;
   			}
		}

	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
    cout<<"matrix_mul order "<<argv[1]<<endl;

    struct timespec start_stl, end_stl;
    if (clock_gettime(CLOCK_MONOTONIC, &start_stl) < 0) {
		perror("clock_gettime");
		return 1;
    }

    serial_dandc(0,n,0,n,0,n,a,n,b,n,c,n,X);

    if (clock_gettime(CLOCK_MONOTONIC, &end_stl) < 0) {
		perror("clock_gettime");
		return 1;
    }

	if (n < 10) 
		print_matrix(c,n);

    struct timespec start, end;
    if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
		perror("clock_gettime");
	return 1;
    }
    parallel_dandc(0,n,0,n,0,n,a,n,b,n,d,n,X);
    if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
		perror("clock_gettime");
	return 1;
    }
    long run_time1 = print_time(std::cout, start, end);
    long run_time =  print_time(std::cout, start_stl, end_stl);
    cout << " cilk paralell Matrix multiply took " << run_time1 << " mseconds."<<endl;
    cout << " cilk serial Matrix Multiply took " << run_time  << " mseconds."<< endl;
	

	cout<<"cilk speedup = "<< (double)run_time/(double)run_time1<<endl;
//	if (n < 10) 
//		print_matrix(d,n);


	for(int i=0;i<n;i++)
        {
        	for(int j=0;j<n;j++)
                         {
                                 assert(c[i*n + j] == d[i*n +j]);
                         }
        }

	free(a);
	free(b);
	free(c);
	free(d);
	}
	return 0;
}


