#include <iostream>
#include <assert.h>
#include <cilk/cilk.h>
#include <cilk/cilk_api.h>
#include <time.h>
using namespace std;

#define BASE 25

long print_time(std::ostream &out, struct timespec start,
                       struct timespec end){
    time_t s = end.tv_sec - start.tv_sec;
    long ns = end.tv_nsec - start.tv_nsec;
    if (ns < 0) {
		ns += 1000000000;
		s -= 1;
    }
    return (s * 1000 + ((ns + 500000) / 1000000));
}

long fib_serial(long n)
{
    if (n < 2) return n;
        return fib_serial(n-1)+fib_serial(n-2);
}

long fib_parallel(long n)
{
    long x, y;
    if (n < 2) 
		return n;
    else if (n < BASE)
		{return fib_serial(n);}
		else
			{
				x = cilk_spawn fib_parallel(n-1);
    				y = fib_parallel(n-2);
    				cilk_sync;
    				return (x+y);
			}
}
long
fib[]={0,1,1,2,3,5,8,13,21,34,55, 89, 144, 233, 377, 610, 987, 1597, 2584, 4181, 6765, 10946, 17711, 28657, 46368, 75025, 121393, 196418, 317811, 514229, 832040, 1346269, 2178309, 3524578, 5702887, 9227465, 14930352, 24157817, 39088169, 63245986, 102334155, 165580141, 267914296, 433494437, 701408733, 1134903170, 1836311903, 2971215073, 4807526976, 7778742049, 12586269025, 20365011074, 32951280099, 53316291173, 86267571272, 139583862445, 225851433717, 365435296162, 591286729879, 956722026041, 1548008755920};


long fib_parallel_check(long n)
{
    long x, y;
    if (n < 2) 
	return n;
    else if (n < BASE)
		{return fib_serial(n);}
		else
			{
				x = cilk_spawn fib_parallel_check(n-1);
    				y = fib_parallel_check(n-2);
    				cilk_sync;
				long re = x+y;
				assert(re == fib[n]); 
    				return re;
			}
}

int main(int argc, char * argv[]) 
{
    long n,s;

	if (argc == 1){
		cout<<"Usage: ./fibo order [check]"<<endl;
        	exit(0);
	}
	else if ( argc >= 2 ){
        	n = atoi(argv[1]);
	}

    cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
	cout<<"fibonacci order "<<argv[1]<<endl;

    struct timespec start, end;
    if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
		perror("clock_gettime");
		return 1;
    }

	if(argc == 3)
 	{
	    fib_parallel_check(n);
	}
	else
	{
		cout << fib_parallel(n) << endl; 

    }

	 if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
		perror("clock_gettime");
		return 1;
    }


        
    struct timespec start_s, end_s;
    if (clock_gettime(CLOCK_MONOTONIC, &start_s) < 0) {
		perror("clock_gettime");
		return 1;
    }

      //  fib_serial(n);
cout<<fib_serial(n)<<endl; 

    if (clock_gettime(CLOCK_MONOTONIC, &end_s) < 0) {
		perror("clock_gettime");
		return 1;
    }

    long run_time1 = print_time(std::cout, start, end);
    long run_time =  print_time(std::cout, start_s, end_s);
    cout << " cilk paralell fibonacci took " << run_time1 << " mseconds."<<endl;
    cout << " cilk serial fibonacci took " << run_time  << " mseconds."<< endl;
	cout<<"cilk speedup = "<< (double)run_time/(double)run_time1<<endl;

	return 0;

}
