/* out-of-place rectangular matrix transposition test */
#include <cilk/cilk.h>
#include <cilk/cilk_api.h>
#include <iostream>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <cilktools/cilkview.h>
using namespace std;
unsigned test_status = 0;
bool verbose = true;
int THRESHOLD;
#define TEST_ASSERT(c) do { if (!(c)) { \
    std::cout << __FILE__ << "(" << __LINE__ << ") : Error: Assert failed: " \
              << #c << std::endl; ++test_status; } } while (false)

template <typename T>
T hash(T i, T j)
{
     return 17 * i + j;
}

/* out-of-place transpose A[i0..i1][j0..j1] into B[j0..j1][i0..i1] */
template <typename T>
void transpose_serial(T *A, int lda, T *B, int ldb,
	       int i0, int i1, int j0, int j1)
{
 tail:
     int di = i1 - i0, dj = j1 - j0;
     if (di >= dj && di > THRESHOLD) {
	  int im = (i0 + i1) / 2;
	  transpose_serial(A, lda, B, ldb, i0, im, j0, j1);
	  i0 = im; goto tail;
     } else if (dj > THRESHOLD) {
	  int jm = (j0 + j1) / 2;
	  transpose_serial(A, lda, B, ldb, i0, i1, j0, jm);
	  j0 = jm; goto tail;
     } else {
	  for (int i = i0; i < i1; ++i)
	       for (int j = j0; j < j1; ++j) 
		    B[j * ldb + i] = A[i * lda + j];
     }
}

template <typename T>
void transpose(T *A, int lda, T *B, int ldb,
	       int i0, int i1, int j0, int j1)
{
 tail:
     int di = i1 - i0, dj = j1 - j0;
     if (di >= dj && di > THRESHOLD) {
	  int im = (i0 + i1) / 2;
	  cilk_spawn transpose(A, lda, B, ldb, i0, im, j0, j1);
	  i0 = im; goto tail;
     } else if (dj > THRESHOLD) {
	  int jm = (j0 + j1) / 2;
	  cilk_spawn transpose(A, lda, B, ldb, i0, i1, j0, jm);
	  j0 = jm; goto tail;
     } else {
	  for (int i = i0; i < i1; ++i)
	       for (int j = j0; j < j1; ++j) 
		    B[j * ldb + i] = A[i * lda + j];
     }
}

/* A has ni rows and nj columns, row-major layout */
template <typename T>
void test(int ni, int nj, const char *type)
{
     int n = ni * nj;
     T *A = new T[n];
     T *B = new T[n];

     int lda = nj;
     int ldb = ni;

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j) {
	       A[i * lda + j] = hash<T>(i, j);
	  }

     double start = __cilkview_getticks();
     transpose(A, lda, B, ldb, 0, ni, 0, nj);
     double end = __cilkview_getticks();

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j)
	       TEST_ASSERT(B[j * ldb + i] == hash<T>(i, j));

	double run_time= end -start;
     /* do not output timing unless requested, or the test comparison
	will fail */
     if (verbose)
	  std::cout << "cilk parallel matrix tran = "<< run_time  << " ms"<<endl;

     start = __cilkview_getticks();
     transpose_serial(A, lda, B, ldb, 0, ni, 0, nj);
     end = __cilkview_getticks();

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j)
	       assert(B[j * ldb + i] == hash<T>(i, j));

	double run_time1 = end - start;
     /* do not output timing unless requested, or the test comparison
	will fail */
     if (verbose)
	  std::cout << "cilk serial matrix tran = "<<run_time1  << " ms"<<endl;;
     cout<<"cilk speedup = "<< run_time1/run_time<<endl;
	delete[] A;
     delete[] B;
}

int cilk_main_(int argc, char *argv[])
{
     int ni = 4577;
     int nj = 4893;

     verbose = (argc > 1);

     if (argc > 1)
	  ni = atoi(argv[1]);
     if (argc > 2)
	  nj = atoi(argv[2]);

//     test<char>     (ni, nj, "char");
//     test<short>    (ni, nj, "short");
     test<int>      (ni, nj, "int");
//     test<long long>(ni, nj, "long long");
//     test<float>    (ni, nj, "float");
//     test<double>   (ni, nj, "double");

     return test_status;
}

struct args_s {
    int argc;
    char **argv;
};

int  cilk_main_0(void *args_)
{
    args_s *args = static_cast<args_s *>(args_);
    return cilk_main_(args->argc, args->argv);
}


int main(int argc, char *argv[])
{
     int nWorkers = 0;
     args_s args;
     char* arg0 = argv[0];
	int ni,nj;
	if(argc==1)
	{
		cout<<"Usage: ./mm_transpose workers order order basecase"<<endl;
		exit(0);
	}
	else
	{
		__cilkrts_set_param("nworkers",argv[1]);
		ni = atoi(argv[2]);
		nj = atoi(argv[3]);
		THRESHOLD = atoi(argv[4]);
	}
	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
	cout<<"matrix_transpose order "<<argv[2]<<endl;
     test<int>      (ni, nj, "int");

}
