/* out-of-place rectangular matrix transposition test */
#include <iostream>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <cilk/cilk.h>
#include <cilk/cilk_api.h>
#include <cilk/cilkscale.h>
#include <time.h>
using namespace std;
unsigned test_status = 0;
bool verbose = true;
int THRESHOLD, DEPTH;
#define TEST_ASSERT(c) do { if (!(c)) { \
    std::cout << __FILE__ << "(" << __LINE__ << ") : Error: Assert failed: " \
              << #c << std::endl; ++test_status; } } while (false)


long print_time(std::ostream &out, struct timespec start,
                       struct timespec end){
    time_t s = end.tv_sec - start.tv_sec;
    long ns = end.tv_nsec - start.tv_nsec;
    if (ns < 0) {
		ns += 1000000000;
		s -= 1;
    }
    return (s * 1000 + ((ns + 500000) / 1000000));
}


template <typename T>
T myhash(T i, T j){
     return 17 * i + j;
}

/* out-of-place transpose A[i0..i1][j0..j1] into B[j0..j1][i0..i1] */
template <typename T>
void transpose_serial(T *A, int lda, T *B, int ldb,
	       int i0, int i1, int j0, int j1)
{
 tail:
     int di = i1 - i0, dj = j1 - j0;
     if (di >= dj && di > THRESHOLD) {
	  int im = (i0 + i1) / 2;
	  transpose_serial(A, lda, B, ldb, i0, im, j0, j1);
	  i0 = im; goto tail;
     } else if (dj > THRESHOLD) {
	  int jm = (j0 + j1) / 2;
	  transpose_serial(A, lda, B, ldb, i0, i1, j0, jm);
	  j0 = jm; goto tail;
     } else {
	  for (int i = i0; i < i1; ++i)
	       for (int j = j0; j < j1; ++j) 
		    B[j * ldb + i] = A[i * lda + j];
     }
}

template <typename T>
void transpose(T *A, int lda, T *B, int ldb,
	       int i0, int i1, int j0, int j1, int depth)
{
 tail:
    int di = i1 - i0, dj = j1 - j0;
    if (di >= dj && di > THRESHOLD) {
		int im = (i0 + i1) / 2;
		if(depth > 0){
			cilk_spawn transpose(A, lda, B, ldb, i0, im, j0, j1,depth - 1);
			transpose(A, lda, B, ldb, im, i1, j0, j1,depth - 1);
		}
		else{
			transpose_serial(A, lda, B, ldb, i0, im, j0, j1);
			i0 = im;
			goto tail;
		}
    } else if (dj > THRESHOLD) {
		int jm = (j0 + j1) / 2;
		if(depth > 0){
			cilk_spawn transpose(A, lda, B, ldb, i0, i1, j0, jm, depth - 1);
			transpose(A, lda, B, ldb, i0, i1, jm, j1, depth - 1);
		}
		else{
			transpose_serial(A, lda, B, ldb, i0, i1, j0, jm);
			j0 = jm;
			goto tail;
		}
    } else {
		for (int i = i0; i < i1; ++i)
	    	 for(int j = j0; j < j1; ++j) 
				B[j * ldb + i] = A[i * lda + j];
    }
} 

/* A has ni rows and nj columns, row-major layout */
template <typename T>
void test(int ni, int nj, int depth, const char *type)
{
     int n = ni * nj;
     T *A = new T[n];
     T *B = new T[n];

     int lda = nj;
     int ldb = ni;

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j) {
	       A[i * lda + j] = myhash<T>(i, j);
	  }

    struct timespec start_stl, end_stl;
    if (clock_gettime(CLOCK_MONOTONIC, &start_stl) < 0) {
		perror("clock_gettime");
		exit(1);
    }
  //  for(int i = 0; i < 100 ; i ++)
    	transpose_serial(A, lda, B, ldb, 0, ni, 0, nj);

     if (clock_gettime(CLOCK_MONOTONIC, &end_stl) < 0){
		perror("clock_gettime");
		exit(1);
    }
         for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j)
	       TEST_ASSERT(B[j * ldb + i] == myhash<T>(i, j));


   	 delete[] A;
     delete[] B;
      A = new T[n];
     B = new T[n];

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j) {
	       A[i * lda + j] = myhash<T>(i, j);
	  }

        struct timespec start, end;
    if (clock_gettime(CLOCK_MONOTONIC, &start) < 0){
		perror("clock_gettime");
		exit(1);
    }
   // for(int i = 0; i < 100 ; i ++)
   // wsp_t scale_start, scale_end;
 //	scale_start = wsp_getworkspan();
    	transpose(A, lda, B, ldb, 0, ni, 0, nj, depth);
    	cilk_sync;
   // 	 scale_end = wsp_getworkspan();
    if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
		perror("clock_gettime");
		exit(1);
    }
  //  wsp_dump(wsp_sub(scale_end, scale_start), "sample_transpose");

     for (int i = 0; i < ni; ++i)
	  for (int j = 0; j < nj; ++j)
	       assert(B[j * ldb + i] == myhash<T>(i, j));


    long run_time1 = print_time(std::cout, start, end);
    long run_time =  print_time(std::cout, start_stl, end_stl);
    cout << " cilk paralell Matrix transpose took " << run_time1 << " mseconds."<<endl;
    cout << " cilk serial Matrix transpose took " << run_time  << " mseconds."<< endl;
	

	cout<<"cilk speedup = "<< (double)run_time/(double)run_time1<<endl;
//	if (n < 10) 
	delete[] A;
     delete[] B;
}



int main(int argc, char *argv[])
{
     int nWorkers = 0;
     char* arg0 = argv[0];
	int ni,nj;
	int depth = 3;
	if(argc==1)
	{
		cout<<"Usage: ./mm_transpose order order basecase depth"<<endl;
		exit(0);
	}
	else
	{
		ni = atoi(argv[1]);
		nj = atoi(argv[2]);
		THRESHOLD = atoi(argv[3]);
		depth = atoi(argv[4]);

	}
	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
	cout<<"matrix_transpose order "<<argv[1] <<" " << argv[2] <<endl;
     test<int>(ni, nj, depth, "int");

}
