#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <malloc.h>
#include <iostream>
#include <assert.h>
#include <cilk/cilk.h>
#include <cilk/cilk_api.h>
#include <time.h>
using namespace std;
#define basecase 4096
long print_time(std::ostream &out, struct timespec start,
                       struct timespec end){
    time_t s = end.tv_sec - start.tv_sec;
    long ns = end.tv_nsec - start.tv_nsec;
    if (ns < 0) {
		ns += 1000000000;
		s -= 1;
    }
    return (s * 1000 + ((ns + 500000) / 1000000));
}
void alloc_data(int* data, int size)
{
	if(size <= 0) 
	{
		printf("size must be > 0\n") ;
		exit(1) ;
	}
	data =(int *)malloc(sizeof(int)*size) ;
	if( data == NULL) 
	{
		printf("Cannot allocate data: %ld bytes\n", size * sizeof(int));
		exit(1) ;
	}
}

void dealloc_data(int* data)
{
	free(data) ;
}


void init(int *a, int n)
{
	for (int i = 0; i < n; i++)
	{
		cout<<"yes"<<&a<<endl;;
	//	a[i] = i;
	}
}


void merge(int* data, int istart, int mid, int iend)
{
	int n = iend - istart + 1 ;
	int k = mid - istart + 1 ;
	int m = n - k ;
	int indexA = 0, indexB = 0, indexC = 0 ;
	int i ;
	int* a = &data[istart] ;
	int* b = &data[mid+1] ;
	int* c = (int *)malloc(sizeof(int)*n);

	while(indexA < k && indexB < m) {
	if(a[indexA] <= b[indexB]) 
	{
		c[indexC] = a[indexA] ;
		indexA++ ;
		indexC++ ;
	} 
	else 
	{
	c[indexC] = b[indexB] ;
	indexB++ ;
	indexC++ ;
	}
	}
	if(indexA >= k) 
	{
		for(i = 0 ; i < n-indexC ; i++) 
		{
			c[indexC+i] = b[indexB+i] ;
		}
	} 
	else 
	{
		for(i = 0 ; i < n-indexC ; i++) 
		{
			c[indexC+i] = a[indexA+i] ;
		}
	}

	for(i = 0 ; i < n ; i++) 
	{
		data[istart+i] = c[i] ;
	}

	free(c) ;
}

/*strict less and equal*/
int binarysearch(int *data, int low, int high, int key)
 {
         int middle;
         int istart = low, iend = high + 1;
         while ( istart < iend )
         {
                  middle = istart + ( iend - istart )/2;
		if ( data[middle] <= key)
			istart = middle + 1;
		else
			iend = middle;
         }
         return istart;

 }

void parallel_submerge(int *c, int *data, int lowx, int highx, int lowy, int highy, int sp)
{

	int k = highx - lowx + 1 ;
	int m = highy - lowy + 1 ;
	int indexA = 0, indexB = 0, indexC = sp ;
	int i ;
	int* a = &data[lowx] ;
	int* b = &data[lowy] ;

	while(indexA < k && indexB < m) {
	if(a[indexA] <= b[indexB]) 
	{
		c[indexC] = a[indexA] ;
		indexA++ ;
		indexC++ ;
	} 
	else 
	{
	c[indexC] = b[indexB] ;
	indexB++ ;
	indexC++ ;
	}
	}
	if(indexA >= k) 
	{
		for(i = 0 ; i < m - indexB ; i++) 
		{
			c[indexC+i] = b[indexB+i] ;
		}
	} 
	else 
	{
		for(i = 0 ; i < k - indexA ; i++) 
		{
			c[indexC+i] = a[indexA+i] ;
		}
	}

}

int parallel_merge(int* c,int* a, int lowx, int highx, int lowy, int highy, int sp)
 {
        int mx,my,p,lx,ly;
	lx = highx - lowx + 1;
	ly = highy - lowy + 1;
        if (ly <= basecase || lx <=basecase)
	{
		parallel_submerge(c, a, lowx, highx, lowy, highy,sp);
	}
	else if (lx < ly)  
               parallel_merge(c, a, lowy, highy, lowx, highx, sp);
         else if ( lx == 0 )
        {
                         return 0;

        }        else
                 {     
                         mx = (lowx + highx)/2;
                         my = binarysearch(a, lowy, highy, a[mx]);
                         p = sp + mx - lowx + my - lowy;
                         c[p]=a[mx];
                         cilk_spawn parallel_merge(c, a, lowx, mx - 1, lowy, my-1, sp);
                         parallel_merge(c, a, mx + 1, highx, my, highy, p + 1);
 }
 return 0;
}



void serial_mergesort(int* data, int istart, int iend)
{
	if(istart < iend) 
	{
		int mid = (istart + iend) / 2 ;
		serial_mergesort(data, istart, mid) ;
		serial_mergesort(data, mid+1, iend) ;
                merge(data, istart, mid, iend) ;
        }
}

void parallel_mergesort(int* c,int* data, int istart, int iend, int BASE)
{
	if (iend - istart <= BASE)
	{	serial_mergesort( data, istart, iend);
//		for(int i=istart; i<=iend;i++)
//			cout<<data[i]<<" ";
//		cout<<endl;
	}
	else if(istart < iend) 
	{
		int mid = (istart + iend) / 2 ;
		cilk_spawn parallel_mergesort(c,data, istart, mid,BASE) ;
		parallel_mergesort(c,data, mid+1, iend,BASE) ;
                cilk_sync;
//		merge(data,istart,mid,iend);	
	parallel_merge(c,data,istart,mid, mid + 1,iend,istart) ;
		for(int i=istart; i<=iend;i++)
			data[i] = c[i];
        }
}



int main(int argc, char* argv[])
{
	double einit ;
	double esort ;
	int* data,*temp1,*temp2,*temp3 ;
	int * c;
	int size ;
        int i,base;
        if (argc == 1)
        {
                cout<<"Usage: ./pmergesort order basecase"<<endl;
                exit(0);
        }
        else if ( argc >= 2 )
        {
                size = atoi(argv[1]);
                base = atoi(argv[2]);
        }
	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
        cout<<"mergesort order "<<argv[1]<<endl;
	srand(time(NULL)) ;
//	alloc_data(data, size) ;

	data =(int *)malloc(sizeof(int)*size) ;
	c =(int *)malloc(sizeof(int)*size) ;
	for(int k=0;k<size;k++)
		data[k]=rand()%100; 


    struct timespec start, end;
    if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
		perror("clock_gettime");
	return 1;
    }
	
	parallel_mergesort(c,data, 0, size-1, base) ;
    if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
		perror("clock_gettime");
	return 1;
    }
	for(int k=0;k<size-1;k++)
		{
		assert(c[k]<=c[k+1]);	
		}
	std::random_shuffle(data,data+size);

	    struct timespec start_stl, end_stl;
    if (clock_gettime(CLOCK_MONOTONIC, &start_stl) < 0) {
		perror("clock_gettime");
		return 1;
    }	serial_mergesort(data, 0, size-1) ;
      
	for(int k=0;k<size-1;k++)
		{
		assert(data[k]<=data[k+1]);	
		}
    if (clock_gettime(CLOCK_MONOTONIC, &end_stl) < 0) {
		perror("clock_gettime");
		return 1;
    }
    long run_time1 = print_time(std::cout, start, end);
    long run_time =  print_time(std::cout, start_stl, end_stl);
    cout << " cilk paralell merge sort took " << run_time1 << " mseconds."<<endl;
    cout << " cilk serial merge sort took " << run_time  << " mseconds."<< endl;
		cout<<"cilk speedup = "<< (double)run_time/(double)run_time1<<endl;
	

	dealloc_data(data) ;
	dealloc_data(c) ;
return 0 ;

}

