#include <algorithm>
#include <iostream>
#include <iterator>
#include <cilk/cilk.h>
#include <cilktools/cilkview.h>
#include <cilk/cilk_api.h>
#include <functional>
using namespace std; 
// Sort the range between bidirectional iterators begin and end.
// end is one past the final element in the range.
// Use the Quick Sort algorithm, using recursive divide and conquer.

void serial_qsort(int * begin, int * end)
{
    if (begin != end) {
        --end;  // Exclude last element (pivot) from partition
        int * middle = std::partition(begin, end,
                          std::bind2nd(std::less<int>(), *end));
        using std::swap;
        swap(*end, *middle);    // move pivot to middle
        serial_qsort(begin, middle);
        serial_qsort(++middle, ++end); // Exclude pivot and restore end
    }
}

void parallel_qsort(int * begin, int * end,int base)
{
	int length;
	length = end - begin;
	if (length < base)
		serial_qsort(begin,end);
	else if (begin != end) {
        --end;  // Exclude last element (pivot) from partition
        int * middle = std::partition(begin, end,
                          std::bind2nd(std::less<int>(), *end));
        using std::swap;
        swap(*end, *middle);    // move pivot to middle
        cilk_spawn parallel_qsort(begin, middle,base);
        parallel_qsort(++middle, ++end,base); // Exclude pivot and restore end
        cilk_sync;
    }
}

void init(int *a, int n)
{
 for (int i = 0; i < n; ++i)
        a[i] = i;
}


int main(int argc, char* argv[])
{
        int n,base;
	if (argc == 1)
        {
                cout<<"Usage: ./qsort workers order basecase"<<endl;
                exit(0);
        }
	else if ( argc > 3 )
        {
                __cilkrts_set_param("nworkers", argv[1]);
        	n = atoi(argv[2]);
		base = atoi(argv[3]);
	}

	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
        cout<<"qsort order "<<argv[2]<<endl;
	int *a = new int[n];
	init(a,n);
        std::random_shuffle(a, a + n);
	double start ,end, run_time, run_time1;
	
	start = __cilkview_getticks();
        parallel_qsort(a, a + n,base);
	end = __cilkview_getticks();
        run_time1 = (end - start);
        cout<<"cilk parallel qsort : " << run_time1 << " mseconds" <<endl;
	for(int k=0;k<n-1;k++)
                {
        	        assert(a[k]<=a[k+1]);
                }

	std::random_shuffle(a, a + n);
	
	start = __cilkview_getticks();

        serial_qsort(a, a + n);
	end = __cilkview_getticks();
        run_time = (end - start);
        cout <<"cilk serial qsort is : " << run_time << " mseconds" <<endl;
        cout<<"cilk speedup :"<<run_time/run_time1<<endl;
        
	for(int k=0;k<n-1;k++)
                {
        	        assert(a[k]<=a[k+1]);
                }
	return 0;
}
