#include <algorithm>
#include <iostream>
#include <iterator>
#include <assert.h>
#include <cilk/cilk.h>
#include <time.h>
#include <cilk/cilk_api.h>
#include <functional>
using namespace std; 
// Sort the range between bidirectional iterators begin and end.
// end is one past the final element in the range.
// Use the Quick Sort algorithm, using recursive divide and conquer.

long print_time(std::ostream &out, struct timespec start,
                       struct timespec end){
    time_t s = end.tv_sec - start.tv_sec;
    long ns = end.tv_nsec - start.tv_nsec;
    if (ns < 0) {
        ns += 1000000000;
        s -= 1;
    }
    return (s * 1000 + ((ns + 500000) / 1000000));
}
void serial_qsort(int * begin, int * end)
{
    if (begin != end) {
        --end;  // Exclude last element (pivot) from partition
        int * middle = std::partition(begin, end,
                          std::bind2nd(std::less<int>(), *end));
        using std::swap;
        swap(*end, *middle);    // move pivot to middle
        serial_qsort(begin, middle);
        serial_qsort(++middle, ++end); // Exclude pivot and restore end
    }
}

void parallel_qsort(int * begin, int * end,int base)
{
	int length;
	length = end - begin;
	if (length < base)
		serial_qsort(begin,end);
	else if (begin != end) {
        --end;  // Exclude last element (pivot) from partition
        int * middle = std::partition(begin, end,
                          std::bind2nd(std::less<int>(), *end));
        using std::swap;
        swap(*end, *middle);    // move pivot to middle
        cilk_spawn parallel_qsort(begin, middle,base);
        parallel_qsort(++middle, ++end,base); // Exclude pivot and restore end
        cilk_sync;
    }
}

void init(int *a, int n)
{
 for (int i = 0; i < n; ++i)
        a[i] = i;
}


int main(int argc, char* argv[])
{
        int n,base;
	if (argc != 3)
        {
                cout<<"Usage: ./qsort order basecase"<<endl;
                exit(0);
        }
	else
        {
        	n = atoi(argv[1]);
		base = atoi(argv[2]);
	}

	cout<<"Using "<<__cilkrts_get_nworkers()<<" workers"<<endl;
        cout<<"qsort order "<<argv[1]<<endl;
	int *a = new int[n];
	init(a,n);
        std::random_shuffle(a, a + n);

	

    struct timespec start, end;
    if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
        perror("clock_gettime");
    return 1;
    }
        parallel_qsort(a, a + n,base);
    if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
        perror("clock_gettime");
    return 1;
    }
	for(int k=0;k<n-1;k++)
                {
        	        assert(a[k]<=a[k+1]);
                }

	std::random_shuffle(a, a + n);
	
    struct timespec start_stl, end_stl;
    if (clock_gettime(CLOCK_MONOTONIC, &start_stl) < 0) {
        perror("clock_gettime");
        return 1;
    }

        serial_qsort(a, a + n);
	 if (clock_gettime(CLOCK_MONOTONIC, &end_stl) < 0) {
        perror("clock_gettime");
        return 1;
    }
 long run_time1 = print_time(std::cout, start, end);
    long run_time =  print_time(std::cout, start_stl, end_stl);
    cout << " cilk paralell qsort took " << run_time1 << " mseconds."<<endl;
    cout << " cilk serial qsort took " << run_time  << " mseconds."<< endl;
    

    cout<<"cilk speedup = "<< (double)run_time/(double)run_time1<<endl;       
	for(int k=0;k<n-1;k++)
                {
        	        assert(a[k]<=a[k+1]);
                }
	return 0;
}
